use crate::monzo::monzo_client_freshened;
use crate::state::{Config, MonzoState, State, StateInner};
use crate::web::warp_main;
use matrix_sdk::room::Room;
use matrix_sdk::ruma::events::room::message::{MessageEventContent, MessageType};
use matrix_sdk::ruma::events::SyncMessageEvent;
use matrix_sdk::ruma::UserId;
use matrix_sdk::{Client, ClientConfig, SyncSettings};
use std::convert::TryFrom;
use std::sync::Arc;
use tokio::io::AsyncReadExt;
use tokio::sync::RwLock;

pub mod monzo;
pub mod state;
pub mod web;

async fn on_room_message(
    event: &SyncMessageEvent<MessageEventContent>,
    room: Room,
    state: &State,
) -> anyhow::Result<()> {
    if let Room::Joined(room) = room {
        if *room.room_id() != state.config.matrix_room {
            return Ok(());
        }
        if let MessageType::Text(body) = &event.content.msgtype {
            if body.body.as_str() == "?" {
                // Respond with our balances.
                if let Some(client) = monzo_client_freshened(state).await? {
                    let mut buf = String::new();
                    let state_inner = state.inner.read().await;
                    for (acc_id, acc_name) in
                        state_inner.monzo_state.as_ref().unwrap().accounts.iter()
                    {
                        let balance = client.balance(acc_id).await?;
                        if !buf.is_empty() {
                            buf.push('\n');
                        }
                        if &balance.currency == "GBP" {
                            let bal_quid = balance.balance / 100;
                            let bal_pennies = balance.balance % 100;
                            buf.push_str(&format!("{}: £{}.{:02}", acc_name, bal_quid, bal_pennies))
                        } else {
                            buf.push_str("?currency");
                        }
                    }
                    let content = MessageEventContent::text_plain(&buf);
                    room.send(content, None).await.unwrap();
                } else {
                    let content = MessageEventContent::text_plain("Not linked :(.");
                    room.send(content, None).await.unwrap();
                }
            }
        }
    }
    Ok(())
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    // Load the config from the environment!
    let config: Config = envy::from_env()?;

    let state_inner = StateInner::default();

    let mxid = UserId::try_from(config.matrix_id.clone())?;
    let store_path = config.matrix_store.clone();

    let client_config = ClientConfig::new().store_path(store_path);

    let client = Client::new_from_user_id_with_config(mxid.clone(), client_config).await?;

    let state = State {
        config: Arc::new(config.clone()),
        inner: Arc::new(RwLock::new(state_inner)),
        matrix_client: Arc::new(client.clone()),
    };

    eprintln!("Logging in!");
    client
        .login(
            mxid.localpart(),
            &config.matrix_password,
            Some("mxmonzo"),
            Some("rei's MxMonzo"),
        )
        .await?;

    eprintln!("Syncing once!");
    client.sync_once(SyncSettings::new()).await?;

    if config.monzo_persist.exists() {
        eprintln!("Loading found monzo persist data.");
        let mut buf = Vec::new();
        tokio::fs::File::open(&config.monzo_persist)
            .await?
            .read_to_end(&mut buf)
            .await?;
        let monzo_state: MonzoState = serde_json::from_slice(&buf)?;
        let mut state_inner = state.inner.write().await;
        state_inner.monzo_state = Some(monzo_state);
    } else {
        eprintln!("(Monzo persist file does not exist.)");
    }

    eprintln!("Starting warp.");
    let _warp_task = tokio::task::spawn(warp_main(state.clone()));

    let state_arc = Arc::new(state);

    client
        .register_event_handler(
            move |ev: SyncMessageEvent<MessageEventContent>, room: Room| {
                let state_arc = state_arc.clone();
                async move {
                    if let Err(error) = on_room_message(&ev, room, &state_arc).await {
                        eprintln!("Error handling message {}: {:?}", ev.event_id, error);
                    }
                }
            },
        )
        .await;

    // Syncing is important to synchronise the client state with the server.
    // This method will never return.
    client
        .sync(
            SyncSettings::default().token(
                client
                    .sync_token()
                    .await
                    .expect("Just synced, should have a token."),
            ),
        )
        .await;

    Ok(())
}
