use crate::state::{persist_monzo_state, MonzoState, State};
use monzo::inner_client::Quick;
use serde::Deserialize;
use std::time::{Duration, SystemTime, UNIX_EPOCH};

#[derive(Deserialize)]
struct MonzoAuthResponse {
    access_token: String,
    expires_in: u64,
    refresh_token: String,
    #[allow(dead_code)]
    client_id: String,
    #[allow(dead_code)]
    token_type: String,
    #[allow(dead_code)]
    user_id: String,
}

pub async fn monzo_client_from_code(
    state: &State,
    code: &str,
) -> anyhow::Result<monzo::Client<monzo::inner_client::Quick>> {
    // obtain access & refresh token from Monzo
    let client = reqwest::Client::new();
    let auth_resp: MonzoAuthResponse = client
        .post("https://api.monzo.com/oauth2/token")
        .form(&[
            ("grant_type", "authorization_code"),
            ("client_id", &state.config.monzo_client_id),
            ("client_secret", &state.config.monzo_client_secret),
            (
                "redirect_uri",
                &format!("{}/auth_done", state.config.base_uri),
            ),
            ("code", code),
        ])
        .send()
        .await?
        .json()
        .await?;

    let mut state_inner = state.inner.write().await;
    state_inner.monzo_state = Some(MonzoState {
        access_token: auth_resp.access_token.clone(),
        access_token_expiry: (SystemTime::now() + Duration::from_secs(auth_resp.expires_in))
            .duration_since(UNIX_EPOCH)?
            .as_secs(),
        refresh_token: auth_resp.refresh_token,
        accounts: Default::default(),
    });

    let client = monzo::Client::new(auth_resp.access_token);

    drop(state_inner);
    persist_monzo_state(state).await?;
    Ok(client)
}

pub async fn monzo_client_freshened(
    state: &State,
) -> anyhow::Result<Option<monzo::Client<monzo::inner_client::Quick>>> {
    let state_inner = state.inner.read().await;
    let monzo_state = match &state_inner.monzo_state {
        None => {
            return Ok(None);
        }
        Some(ms) => ms,
    };

    // Allow for 2 minute buffer.
    let exp_at_instant = UNIX_EPOCH + Duration::from_secs(monzo_state.access_token_expiry - 120);
    let refresh_token = monzo_state.refresh_token.clone();
    drop(state_inner);
    let needs_refresh = exp_at_instant < SystemTime::now();
    if needs_refresh {
        eprintln!("Refreshing token; expiry {:?}; now {:?}.", exp_at_instant, SystemTime::now());
        eprintln!("DEBUG refresh_token being used {:?}", &refresh_token);
        let client = reqwest::Client::new();
        let resp = client
            .post("https://api.monzo.com/oauth2/token")
            .form(&[
                ("grant_type", "refresh_token"),
                ("client_id", &state.config.monzo_client_id),
                ("client_secret", &state.config.monzo_client_secret),
                ("refresh_token", &refresh_token),
            ])
            .send()
            .await?;
        eprintln!("DEBUG response status {:?}", resp.status());
        let jv: serde_json::Value = resp.json()
            .await?;
        eprintln!("DEBUG raw JV {:#?}", jv);
        let auth_resp: MonzoAuthResponse = serde_json::from_value(jv)?;
        let mut state_inner = state.inner.write().await;
        let monzo_state = state_inner.monzo_state.as_mut().unwrap();
        monzo_state.access_token = auth_resp.access_token;
        monzo_state.refresh_token = auth_resp.refresh_token;
        monzo_state.access_token_expiry = (SystemTime::now()
            + Duration::from_secs(auth_resp.expires_in))
        .duration_since(UNIX_EPOCH)?
        .as_secs();
    }

    let state_inner = state.inner.read().await;
    let monzo_state = state_inner.monzo_state.as_ref().unwrap();

    let client = monzo::Client::new(monzo_state.access_token.clone());

    if needs_refresh {
        persist_monzo_state(state).await?;
    }
    Ok(Some(client))
}

#[derive(Deserialize)]
pub struct Webhook {
    pub account_id: String,
    pub id: String,
    pub url: String,
}

#[derive(Deserialize)]
pub struct WebhookList {
    pub webhooks: Vec<Webhook>,
}

pub async fn list_webhooks(
    client: &monzo::Client<Quick>,
    account_id: &str,
) -> anyhow::Result<Vec<Webhook>> {
    let at = client.access_token();
    let client = reqwest::Client::new();
    let response: WebhookList = client
        .get("https://api.monzo.com/webhooks")
        .query(&[("account_id", account_id)])
        .header("Authorization", format!("Bearer {}", at))
        .send()
        .await?
        .json()
        .await?;
    Ok(response.webhooks)
}

pub async fn delete_webhook(client: &monzo::Client<Quick>, webhook_id: &str) -> anyhow::Result<()> {
    let at = client.access_token();
    let client = reqwest::Client::new();
    if !client
        .delete(format!("https://api.monzo.com/webhooks/{}", webhook_id))
        .header("Authorization", format!("Bearer {}", at))
        .send()
        .await?
        .status()
        .is_success()
    {
        anyhow::bail!("Not successful: delete");
    }
    Ok(())
}

pub async fn create_webhook(
    client: &monzo::Client<Quick>,
    account_id: &str,
    url: &str,
) -> anyhow::Result<()> {
    let at = client.access_token();
    let client = reqwest::Client::new();
    if !client
        .post("https://api.monzo.com/webhooks")
        .header("Authorization", format!("Bearer {}", at))
        .form(&[("account_id", account_id), ("url", url)])
        .send()
        .await?
        .status()
        .is_success()
    {
        anyhow::bail!("Not successful: register hook");
    }
    Ok(())
}
