use crate::monzo::{
    create_webhook, delete_webhook, list_webhooks, monzo_client_freshened, monzo_client_from_code,
};
use crate::state::State;
use matrix_sdk::room::Room;
use matrix_sdk::ruma::events::room::message::MessageEventContent;
use matrix_sdk::ruma::RoomId;
use monzo::accounts::Type;
use serde::Deserialize;
use std::collections::HashMap;
use std::net::SocketAddr;
use std::str::FromStr;
use url::Url;
use warp::http::{Response, StatusCode};
use warp::Filter;

fn wrap<T: warp::Reply + 'static>(
    ok_or_err: anyhow::Result<T>,
) -> Result<Box<dyn warp::Reply>, warp::Rejection> {
    match ok_or_err {
        Ok(inner) => Ok(Box::new(inner)),
        Err(err) => {
            eprintln!("Error handling request {:?}", err);
            Ok(Box::new(warp::reply::with_status(
                "An error has occurred.",
                StatusCode::INTERNAL_SERVER_ERROR,
            )))
        }
    }
}

async fn auth_setup(
    /* host: String, forwarded_proto: Option<String>, */ state: State,
) -> Result<impl warp::Reply, warp::Rejection> {
    // let final_redirect_uri = format!("{}://{}/auth_done", forwarded_proto.as_deref().unwrap_or("http"), host);
    let final_redirect_uri = format!("{}/auth_done", state.config.base_uri);
    let mut url = Url::parse("https://auth.monzo.com/").unwrap();
    url.query_pairs_mut()
        .append_pair("client_id", &state.config.monzo_client_id)
        .append_pair("redirect_uri", &final_redirect_uri)
        .append_pair("response_type", "code");
    Ok(Response::builder()
        .status(StatusCode::SEE_OTHER)
        .header("Location", url.as_str())
        .body("Redirecting to log in..."))
}

#[derive(Deserialize)]
struct AuthDone {
    code: String,
}

async fn auth_done(query: AuthDone, state: State) -> anyhow::Result<impl warp::Reply> {
    eprintln!("code {}", query.code);

    let _client = monzo_client_from_code(&state, &query.code).await?;

    Ok("Success! Please authorise in Monzo, then click <a href='/auth_confirmed'>here</a>.")
}

async fn auth_done_wrapped(
    query: AuthDone,
    state: State,
) -> Result<impl warp::Reply, warp::Rejection> {
    wrap(auth_done(query, state).await)
}

async fn auth_confirmed(state: State) -> anyhow::Result<impl warp::Reply> {
    let client = monzo_client_freshened(&state)
        .await?
        .ok_or_else(|| anyhow::anyhow!("Monzo client has not been set up yet."))?;
    let accounts = client.accounts().await?;

    let mut account_names = HashMap::new();

    for account in accounts {
        let acc_type = match account.account_type {
            Type::UkBusiness => "business",
            Type::UkRetail => "personal",
            Type::UkRetailJoint => "joint",
            _ => "?type",
        };

        // create/overwrite webhooks
        for hook in list_webhooks(&client, &account.id).await? {
            delete_webhook(&client, &hook.id).await?;
        }
        create_webhook(
            &client,
            &account.id,
            &format!("{}/hook", state.config.base_uri),
        )
        .await?;

        account_names.insert(account.id, acc_type.to_string());
    }

    let mut state_inner = state.inner.write().await;
    state_inner.monzo_state.as_mut().unwrap().accounts = account_names;

    Ok("Success!")
}

async fn auth_confirmed_wrapped(state: State) -> Result<impl warp::Reply, warp::Rejection> {
    wrap(auth_confirmed(state).await)
}

#[derive(Deserialize)]
struct MonzoHook {
    #[serde(rename = "type")]
    kind: String,
    #[serde(rename = "data")]
    payload: serde_json::Value,
}

#[derive(Deserialize)]
struct MonzoTransactionCreated {
    amount: i64,
    merchant: Option<Merchant>,
    account_id: String,
    counterparty: Option<Counterparty>,
}

#[derive(Deserialize)]
struct Merchant {
    name: String,
}

#[derive(Deserialize)]
struct Counterparty {
    name: Option<String>,
}

async fn monzo_hook(hook: MonzoHook, state: State) -> anyhow::Result<impl warp::Reply> {
    if &hook.kind == "transaction.created" {
        eprintln!("TODO transaction hook (pre-processed)");
        eprintln!("---\n{:#?}\n---", hook.payload);
        let txn_created: MonzoTransactionCreated = serde_json::from_value(hook.payload)?;
        eprintln!("TODO transaction hook.");

        let room = state
            .matrix_client
            .get_room(&RoomId::from_str(&state.config.matrix_room)?)
            .ok_or_else(|| anyhow::anyhow!("Matrix room not found!"))?;
        if let Room::Joined(room) = room {
            let mut buf = String::new();
            // − +
            if txn_created.amount < 0 {
                buf.push('−');
            } else {
                buf.push('+');
            }
            let quid = txn_created.amount.abs() / 100;
            let pennies = txn_created.amount.abs() % 100;
            buf.push_str(&format!("£{}.{:02} (", quid, pennies));

            if let Some(merchant) = txn_created.merchant {
                buf.push_str(&merchant.name);
            } else if let Some(counterparty) = txn_created.counterparty {
                if let Some(name) = counterparty.name {
                    buf.push_str(&name);
                } else {
                    buf.push_str("?c?");
                }
            } else {
                buf.push_str("???");
            }
            buf.push(')');

            let state_inner = state.inner.read().await;
            let on_account_name = state_inner
                .monzo_state
                .as_ref()
                .map(|x| x.accounts.get(&txn_created.account_id))
                .flatten()
                .map(String::as_str)
                .unwrap_or("unk");

            buf.push_str(&format!(" on {}", on_account_name));
            if let Some(monzo_client) = monzo_client_freshened(&state).await? {
                // include new balance
                match monzo_client.balance(&txn_created.account_id).await {
                    Ok(resp) => {
                        let new_bal = resp.balance;
                        let now_quid = new_bal / 100;
                        let now_pennies = new_bal % 100;
                        buf.push_str(&format!(" now £{}.{:02}", now_quid, now_pennies));
                    }
                    Err(err) => {
                        buf.push_str(". Error when getting now.");
                        eprintln!("when querying balance after webhook: {:?}", err);
                    }
                }

            }

            let content = MessageEventContent::text_plain(&buf);
            room.send(content, None).await.unwrap();
        } else {
            eprintln!("Not in room.");
        }
    } else {
        eprintln!("Unknown webhook kind: {}", hook.kind);
    }
    Ok("")
}

async fn monzo_hook_wrapped(
    body: MonzoHook,
    state: State,
) -> Result<impl warp::Reply, warp::Rejection> {
    wrap(monzo_hook(body, state).await)
}

/// Entry point for the warp-based HTTP server component.
/// What it's for:
/// - setting up Monzo access tokens
/// - receiving Monzo webhooks
pub async fn warp_main(state: State) {
    let with_state = {
        let state = state.clone();
        warp::any().map(move || state.clone())
    };
    let auth_setup_route = warp::get()
        .and(warp::path("auth_setup"))
        // .and(warp::header("Host"))
        // .and(warp::header::optional("X-Forwarded-Proto"))
        .and(with_state.clone())
        .and_then(auth_setup);

    let auth_done_route = warp::get()
        .and(warp::path("auth_done"))
        .and(warp::query())
        .and(with_state.clone())
        .and_then(auth_done_wrapped);

    let auth_confirmed_route = warp::get()
        .and(warp::path("auth_confirmed"))
        .and(with_state.clone())
        .and_then(auth_confirmed_wrapped);

    let hook_route = warp::post()
        .and(warp::path("hook"))
        .and(warp::body::json())
        .and(with_state.clone())
        .and_then(monzo_hook_wrapped);

    let routes = auth_setup_route
        .or(auth_done_route)
        .or(auth_confirmed_route)
        .or(hook_route);

    let address: SocketAddr =
        SocketAddr::from_str(&state.config.bind_address).expect("Failed to parse bind address");
    warp::serve(routes).run(address).await;
}
